`include "vme64x_bfm.svh"
`include "svec_vme_buffers.svh"
`include "fmc.svh"
`include "../../../ffpg/sim/testbench/ffpg_csr.svh"
`include "carrier_csr.svh"
`include "ConfigSvec.svh"

module Testbench;

    reg Reset_inr = 0;
    reg Clk20_ik = 0;
    reg ClkRf_k = 0;
    reg Trigger = 0;
    int BunchCounter = 0;

    always #25ns Clk20_ik <= ~Clk20_ik;
    always #5ns ClkRf_k <= ~ClkRf_k; // 200 MHz
    
    always_ff @(posedge ClkRf_k) begin
        if (BunchCounter == 17820) begin
            BunchCounter <= 0;
            Trigger <= 1;
        end else begin
            BunchCounter <= BunchCounter + 1;
            Trigger <= 0;
        end
    end

    initial begin
        #10us; // for PLL lock and reset settle
        repeat(20) @(posedge Clk20_ik);
            Reset_inr = 1;
    end

    IVME64X VME(Reset_inr);

    `DECLARE_VME_BUFFERS(VME.slave);

    `DECLARE_FMC(0);
    
    logic Fmc1Present_in;

    logic [1:0] FpLedsLineEnable_ob2;
    logic [1:0] FpLedsLine_ob2;
    logic [3:0] FpLedsColumn_ob4;
    wire        CarrierScl_io;
    wire        CarrierSda_io;
    wire        CarrierOneWire_io;

    CBusAccessor_VME64x acc;

    SvecTopFfpgWrapper cDut (
        .Clk20_ik(Clk20_ik),
        .Reset_inr(Reset_inr),
        // Pll20DacDin_o: out std_logic;
        // Pll20DacSclk_o: out std_logic;
        // Pll20DacSync_on: out std_logic;
        // Pll25DacDin_o: out std_logic;
        // Pll25DacSclk_o: out std_logic;
        // Pll25DacSync_on: out std_logic;
        .FpLedsLineEnable_ob2(FpLedsLineEnable_ob2),
        .FpLedsLine_ob2(FpLedsLine_ob2),
        .FpLedsColumn_ob4(FpLedsColumn_ob4),
        .CarrierScl_io(CarrierScl_io),
        .CarrierSda_io(CarrierSda_io),
        .PcbRev_i(5'b00001),
        .CarrierOneWire_io(CarrierOneWire_io),

        `WIRE_FMC(0)
        
        .Fmc1Present_in(Fmc1Present_in),

        `WIRE_VME_PINS(8) // slot number in parameter // don't change that magic 8 - it's hardcoded somewhere else
    );
    
    assign Fmc0Present_in = 1;
    assign Fmc1Present_in = 0;
    assign Fmc0ClkIn0P_ik = ClkRf_k;
    assign Fmc0ClkIn0N_ik = ~Fmc0ClkIn0P_ik;
    assign Fmc0TriggerP_i = Trigger;
    assign Fmc0TriggerN_i = ~Fmc0TriggerP_i;
    assign Fmc0SpiAd9512Miso_i = Fmc0SpiAd9512Mosi_o; // loopback

    task automatic init_vme64x_core;
        /* map func0 to 0x80000000, A32 */

        acc.write('h7ff63, 'h80, A32|CR_CSR|D08Byte3);
        acc.write('h7ff67, 0, CR_CSR|A32|D08Byte3);
        acc.write('h7ff6b, 0, CR_CSR|A32|D08Byte3);
        acc.write('h7ff6f, 36, CR_CSR|A32|D08Byte3);
        acc.write('h7ff33, 1, CR_CSR|A32|D08Byte3);
        acc.write('h7fffb, 'h10, CR_CSR|A32|D08Byte3); /* enable module (BIT_SET = 0x10) */

        acc.set_default_modifiers(A32 | D32 | SINGLE);
    endtask

    task automatic WbWrite(uint32_t address, uint32_t data, string label = "");
        uint64_t data64 = data;
        
        if (label != "")
            $display(label);
        acc.write(address, data64, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, data);
    endtask
    
    task automatic TestSdb;
        uint64_t d;
        uint32_t base;
        uint32_t address;
        int i;
        
        base = `WB_SVEC_VME_ADDRESS;
        $display("Carrier SDB:");
        for (i=0; i<64; i = i+4) begin
            address = base + i;
            acc.read(address, d, A32|SINGLE|D32);
            $display("[0x%x]: 0x%x", address, d[31:0]);
        end

        base = `WB_FFPG_BASE + 64*0;
        $display("FFPG SDB:");
        for (i=0; i<64; i = i+4) begin
            address = base + i;
            acc.read(address, d, A32|SINGLE|D32);
            $display("[0x%x]: 0x%x", address, d[31:0]);
        end
    endtask
    
    task automatic TestFfpgCsr;
        uint64_t d;
        uint32_t base;
        uint32_t address;

        address = `WB_FFPG_CSR_REG(`ADDR_FFPG_STATUS);
        $display("FFPG Status reg.:");
        acc.read(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);

        address = `WB_FFPG_CSR_REG(`ADDR_FFPG_CONTROL);
        $display("Control reg. write:");
        d = 4;
        acc.write(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);

        address = `WB_FFPG_CSR_REG(`ADDR_FFPG_STATUS);
        $display("FFPG Status reg.:");
        acc.read(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);
    endtask
    
    task automatic TestSpi;
        uint64_t d;
        uint32_t base;
        uint32_t address;

        address = `WB_FFPG_SPI_REG(`SPI_CTRL);
        $display("Control reg. write:");
        d = 'h2618; // ASS, NEG_TX/RX, LEN=24
        acc.write(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);

        address = `WB_FFPG_SPI_REG(`SPI_DIVIDER);
        $display("Divider reg. write:");
        d = 2; // f_SCLK ~ 20.833 MHz
        acc.write(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);

        address = `WB_FFPG_SPI_REG(`SPI_SS);
        $display("SlaveSelect reg. write:");
        d = 1;
        acc.write(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);

        address = `WB_FFPG_SPI_REG(`SPI_TX_RX_0);
        $display("TX reg. write:");
        d = 'h80A500; // READ register 0xA5
        acc.write(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);

        address = `WB_FFPG_SPI_REG(`SPI_CTRL);
        $display("Control reg. write:");
        d = 'h2718; // ASS, NEG_TX/RX, LEN=24 + GO
        acc.write(address, d, A32|SINGLE|D32);
        $display("[0x%x]: 0x%x", address, d[31:0]);
        
        #2us;

    endtask

    task automatic TestReal;
        // test on CH2
        int i;
        
        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_CONTROL), 0);
        
        WbWrite(`WB_FFPG_SPI_REG(`SPI_SS), 1);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_DIVIDER), 2);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2618);
        
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h3f03);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4505);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4a00);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4c00);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4e00);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5000);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5200);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4b00);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4d00);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4f00);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5100);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5300);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5820);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5a01);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        
        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_TRIGGER_THRESHOLD), 'h1999);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h4505);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_TX_RX_0), 'h5a01);
        WbWrite(`WB_FFPG_SPI_REG(`SPI_CTRL), 'h2718);

        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_VCXO_VOLTAGE), 'h42b2);

        for (i=0; i<2048*4; i = i+4) begin
            WbWrite(`WB_FFPG_CSR_REG(`BASE_FFPG_CH2_SET_MEM)+i, 0);
            WbWrite(`WB_FFPG_CSR_REG(`BASE_FFPG_CH2_RES_MEM)+i, 0);
        end
        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_OVERFLOW), 'h459c);
        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_TRIGGER_LATENCY), 'ha);

        WbWrite(`WB_FFPG_CSR_REG(`BASE_FFPG_CH2_SET_MEM), 'h02000020);
        WbWrite(`WB_FFPG_CSR_REG(`BASE_FFPG_CH2_RES_MEM), 'h40000400);
        WbWrite(`WB_FFPG_CSR_REG(`BASE_FFPG_CH2_SET_MEM)+3*4, 'h20000000);
        WbWrite(`WB_FFPG_CSR_REG(`BASE_FFPG_CH2_RES_MEM)+4*4, 'h4);

        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_CH2_DELAY_SET), 'h64);
        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_CH2_DELAY_RESET), 'hc8);

        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_CONTROL), 8);
        WbWrite(`WB_FFPG_CSR_REG(`ADDR_FFPG_CONTROL), 'h48);
        
        #1ms;
    endtask
    
    initial begin
        acc = new(VME);

        #10us; // for PLL lock and reset settle
        #20us;
        init_vme64x_core;

        // TestSdb;
        // TestFfpgCsr;
        // TestSpi;
        TestReal;

        
        // uint64_t blt_addr[];
        // uint64_t blt_data[];
        // $display("Write data to DDR in BLT\n");
        // blt_addr = {'h3000};
        // blt_data = {'h1, 'h2, 'h3, 'h4, 'h5, 'h6, 'h7, 'h8 ,'h9, 'hA};
        // acc.writem(blt_addr, blt_data, A32|BLT|D32, result);

        // $display("Read data from DDR in BLT");
        // blt_data = {};
        // acc.readm(blt_addr, blt_data, A32|BLT|D32, result);

        // for(i=0; i<10; i++) begin
            // $display("Data %d: 0x%x\n", i, blt_data[i]);
        // end

        $stop;
    end

endmodule
